@@ -10,21 +10,24 @@ package MCE::Signal;
 use strict;
 use warnings;
 
-use Fcntl qw( :flock );
+use Fcntl qw( :flock O_RDONLY );
 use base qw( Exporter );
 
-require File::Path;
+our $VERSION = '1.499_001'; $VERSION = eval $VERSION;
 
 our ($has_threads, $main_proc_id, $prog_name);
+our ($display_die_with_localtime, $display_warn_with_localtime);
 
 BEGIN {
-   $main_proc_id = $$; $prog_name = $0; $prog_name =~ s{^.*[\\/]}{}g;
+   require File::Path;
+
+   $main_proc_id =  $$;
+   $prog_name    =  $0;
+   $prog_name    =~ s{^.*[\\/]}{}g;
 }
 
-our $VERSION = '1.202_001';
-$VERSION = eval $VERSION;
-
 our $tmp_dir = undef;
+
 our @EXPORT_OK = qw( $tmp_dir sys_cmd stop_and_exit );
 our %EXPORT_TAGS = (
    all     => \@EXPORT_OK,
@@ -38,10 +41,12 @@ our %EXPORT_TAGS = (
 ###############################################################################
 
 sub _croak { require Carp; goto &Carp::croak; }
-sub _usage { _croak 'usage: [-use_dev_shm] [-keep_tmp_dir]'; }
+sub _usage { _croak "MCE::Signal error: $_[0] is not a valid option"; }
 sub _flag  { 1; }
 
 my $_keep_tmp_dir = 0;
+my $_no_sigmsg    = 0;
+my $_no_kill9     = 0;
 my $_loaded;
 
 sub import {
@@ -49,19 +54,36 @@ sub import {
    my $class = shift;
    return if ($_loaded++);
 
-   my $_use_dev_shm = 0;
    my @_export_args = ();
+   my $_no_setpgrp  = 0;
+   my $_setpgrp     = 0;
+   my $_use_dev_shm = 0;
 
    while (my $_arg = shift) {
-      $_use_dev_shm  = _flag() and next if ($_arg eq '-use_dev_shm');
+
       $_keep_tmp_dir = _flag() and next if ($_arg eq '-keep_tmp_dir');
-      _usage() if ($_arg =~ /^-/);
+      $_no_sigmsg    = _flag() and next if ($_arg eq '-no_sigmsg');
+      $_no_kill9     = _flag() and next if ($_arg eq '-no_kill9');
+
+      $_no_setpgrp   = _flag() and next if ($_arg eq '-no_setpgrp');
+      $_setpgrp      = _flag() and next if ($_arg eq '-setpgrp');
+
+      $_use_dev_shm  = _flag() and next if ($_arg eq '-use_dev_shm');
+
+      _usage($_arg) if ($_arg =~ /^-/);
       push @_export_args, $_arg;
    }
 
    local $Exporter::ExportLevel = 1;
    Exporter::import($class, @_export_args);
 
+ # ## MCE no longer calls setpgrp by default as of MCE 1.405.
+ # ## Sets the current process group for the current process.
+ # setpgrp(0,0) if ($_no_setpgrp == 0 && $^O ne 'MSWin32');
+
+   ## Sets the current process group for the current process.
+   setpgrp(0,0) if ($_setpgrp == 1 && $^O ne 'MSWin32');
+
    my ($_tmp_dir_base, $_count);
 
    if (exists $ENV{TEMP}) {
@@ -81,7 +103,7 @@ sub import {
    _croak("MCE::Signal::import: '$_tmp_dir_base' is not writeable")
       unless (-w $_tmp_dir_base);
 
-   $_count = 0;
+   $_count  = 0;
    $tmp_dir = "$_tmp_dir_base/$prog_name.$$.$_count";
 
    while ( !(mkdir "$tmp_dir", 0770) ) {
@@ -95,6 +117,8 @@ sub import {
 ##
 ###############################################################################
 
+our ($mce_sess_dir_ref, $mce_spawned_ref);
+
 ## Set traps to catch signals.
 $SIG{XCPU} = \&stop_and_exit if (exists $SIG{XCPU});   ## UNIX SIG 24
 $SIG{XFSZ} = \&stop_and_exit if (exists $SIG{XFSZ});   ## UNIX SIG 25
@@ -117,21 +141,23 @@ $SIG{CHLD} = 'DEFAULT' if ($^O ne 'MSWin32');
 ##
 ###############################################################################
 
-our $mce_spawned_ref = undef;
-
 END {
    my $_exit_status = $?;
-   MCE::Signal->_shutdown_mce($_exit_status);
-   MCE::Signal->stop_and_exit($_exit_status) if ($$ == $main_proc_id);
+
+   MCE::Signal->_shutdown_mce($_exit_status)
+      if (defined $mce_spawned_ref);
+
+   MCE::Signal->stop_and_exit($_exit_status)
+      if ($$ == $main_proc_id);
 }
 
 ###############################################################################
 ## ----------------------------------------------------------------------------
 ## Run command via the system(...) function.
 ##
-## The system function in Perl ignores SIGINT and SIGQUIT.  These 2 signals
+## The system function in Perl ignores SIGINT and SIGQUIT. These 2 signals
 ## are sent to the command being executed via system() but not back to
-## the underlying Perl script.  The code below will ensure the Perl script
+## the underlying Perl script. The code below will ensure the Perl script
 ## receives the same signal in order to raise an exception immediately
 ## after the system call.
 ##
@@ -180,6 +206,7 @@ sub sys_cmd {
       my $_is_sig      = 0;
 
       if (exists $_sig_name_lkup{$_sig_name}) {
+         $mce_spawned_ref = undef;
          $SIG{$_sig_name} = sub { };
          $_exit_status = $_is_sig = 1;
       }
@@ -187,9 +214,8 @@ sub sys_cmd {
          $_exit_status = $_sig_name if ($_sig_name ne '0');
       }
 
-      $SIG{TERM}     = sub { } if ($_sig_name ne 'TERM');
-      $SIG{__DIE__}  = sub { };
-      $SIG{__WARN__} = sub { };
+      $SIG{TERM} = sub { } if ($_sig_name ne 'TERM');
+      $SIG{__DIE__} = $SIG{__WARN__} = sub { };
 
       ## ----------------------------------------------------------------------
 
@@ -216,12 +242,17 @@ sub sys_cmd {
                elsif ($_sig_name eq 'TERM' && -f "$tmp_dir/died") {
                   $_err_msg = "caught signal '__DIE__', exiting";
                }
+               elsif ($_sig_name eq '__DIE__') {
+                  $_err_msg = "caught signal '__DIE__', exiting";
+               }
                elsif ($_sig_name ne 'PIPE') {
                   $_err_msg = "caught signal '$_sig_name', exiting";
                }
 
                ## Display error message.
-               print STDERR "\n## $prog_name: $_err_msg\n" if ($_err_msg);
+               print STDERR "\n## $prog_name: $_err_msg\n"
+                  if ($_err_msg && $_no_sigmsg == 0);
+
                open my $_FH, "> $tmp_dir/killed"; close $_FH;
 
                ## Signal process group to terminate.
@@ -235,6 +266,14 @@ sub sys_cmd {
 
             ## Remove temp directory.
             if (defined $tmp_dir && $tmp_dir ne '' && -d $tmp_dir) {
+
+               if (defined $mce_sess_dir_ref) {
+                  foreach my $_sess_dir (keys %{ $mce_sess_dir_ref }) {
+                     File::Path::rmtree($_sess_dir);
+                     delete $mce_sess_dir_ref->{$_sess_dir};
+                  }
+               }
+
                if ($_keep_tmp_dir == 1) {
                   print STDERR "$prog_name: saved tmp_dir = $tmp_dir\n";
                }
@@ -243,13 +282,17 @@ sub sys_cmd {
                      File::Path::rmtree($tmp_dir);
                   }
                }
+
                $tmp_dir = undef;
             }
 
             ## Signal process group to die.
             if ($_is_sig == 1) {
-               print STDERR "\n" if ($_sig_name ne 'PIPE');
-               kill('KILL', -$$, $main_proc_id);
+               print STDERR "\n"
+                  if ($_sig_name ne 'PIPE' && $_no_sigmsg == 0);
+
+               kill('KILL', -$$, $main_proc_id)
+                  if ($_sig_name eq 'PIPE' || $_no_kill9 == 0);
             }
          }
       }
@@ -257,7 +300,7 @@ sub sys_cmd {
       ## ----------------------------------------------------------------------
 
       ## For child processes.
-      if ($$ != $main_proc_id && $_is_sig == 1) {
+      if ($$ != $main_proc_id && $_is_sig == 1 && -d $tmp_dir) {
 
          ## Obtain lock.
          open my $CHILD_LOCK, '+>>', "$tmp_dir/child.lock";
@@ -269,7 +312,7 @@ sub sys_cmd {
          if ($_handler_cnt == 1) {
 
             ## Notify the main process that I've died.
-            if ($_sig_name eq '__DIE__') {
+            if ($_sig_name eq '__DIE__' && ! -f "$tmp_dir/died") {
                open my $_FH, "> $tmp_dir/died"; close $_FH;
             }
 
@@ -288,8 +331,8 @@ sub sys_cmd {
       ## ----------------------------------------------------------------------
 
       ## Exit thread/process with status.
-      if ($_is_sig == 1) {
-         select(undef, undef, undef, 0.133) for (1..3);
+      if ($_is_sig == 1 && $_no_kill9 == 0) {
+         select(undef, undef, undef, 0.066) for (1..6);
       }
 
       threads->exit($_exit_status) if ($has_threads && threads->can('exit'));
@@ -336,10 +379,24 @@ sub _die_handler {
 
    shift @_ if (defined $_[0] && $_[0] eq 'MCE::Signal');
 
-   local $SIG{__DIE__} = sub { };
-   local $\ = undef; print STDERR $_[0];
+   CORE::die(@_) if $^S;      ## Direct to CORE::die if executing an eval
+
+   local $\ = undef;
+
+   ## Set $MCE::Signal::display_die_with_localtime = 1;
+   ## when wanting the output to contain the localtime.
+
+   if ($MCE::Signal::display_die_with_localtime) {
+      my $_time_stamp = localtime();
+      print STDERR "## $_time_stamp: $prog_name: ERROR:\n", $_[0];
+   }
+   else {
+      print STDERR $_[0];
+   }
 
    MCE::Signal->stop_and_exit('__DIE__');
+
+   CORE::exit;
 }
 
 sub _warn_handler {
@@ -349,12 +406,23 @@ sub _warn_handler {
    ## Ignore thread warnings during exiting.
 
    return if (
-      $_[0] =~ /^Attempt to free unreferenced scalar/            ||
       $_[0] =~ /^A thread exited while \d+ threads were running/ ||
+      $_[0] =~ /^Attempt to free unreferenced scalar/            ||
       $_[0] =~ /^Perl exited with active threads/
    );
 
-   local $\ = undef; print STDERR $_[0];
+   local $\ = undef;
+
+   ## Set $MCE::Signal::display_warn_with_localtime = 1;
+   ## when wanting the output to contain the localtime.
+
+   if ($MCE::Signal::display_warn_with_localtime) {
+      my $_time_stamp = localtime();
+      print STDERR "## $_time_stamp: $prog_name: WARNING:\n", $_[0];
+   }
+   else {
+      print STDERR $_[0];
+   }
 }
 
 1;
@@ -373,11 +441,14 @@ MCE::Signal - Provides tmp_dir creation & signal handling for Many-Core Engine.
 
 =head1 VERSION
 
-This document describes MCE::Signal version 1.202_001
+This document describes MCE::Signal version 1.499_001
 
 =head1 SYNOPSIS
 
- use MCE::Signal qw( [-use_dev_shm] [-keep_tmp_dir] );
+ use MCE::Signal qw( [-keep_tmp_dir] [-use_dev_shm] );
+
+ use MCE;   ## MCE loads MCE::Signal when not present.
+            ## Include MCE::Signal first for options to take effect.
 
 =head1 DESCRIPTION
 
@@ -390,12 +461,39 @@ and terminates itself.
 The location of temp dir resides under $ENV{TEMP} if defined, otherwise
 /dev/shm if writeable and -use_dev_shm is specified, or /tmp.
 
-The temp dir resides under $ENV{TEMP}/mce/ when running Perl on Microsoft
+The temp dir resides under $ENV{TEMP}/mce/ for native Perl on Microsoft
 Windows.
 
-Nothing is exported by default. Exportable are 1 variable and 2 subroutines:
+As of MCE 1.405, MCE::Signal no longer calls setpgrp by default. Pass the
+-setpgrp option to MCE::Signal to call setpgrp.
 
- $tmp_dir          - Path to temporary directory.
+ ## Running MCE through Daemon::Control requires setpgrp to be called.
+
+ use MCE::Signal qw(-setpgrp);
+ use MCE;
+
+The following are available arguments and their meanings.
+
+ -keep_tmp_dir     - The temporary directory is not removed during exiting
+                     A message is displayed with the location afterwards
+
+ -use_dev_shm      - Create the temporary directory under /dev/shm
+
+ -no_sigmsg        - Do not display a message when receiving a signal
+ -no_kill9         - Do not kill -9 after receiving a signal to terminate
+
+ -setpgrp          - Calls setpgrp to set the process group for the process
+
+                     Specify this option to ensure all workers terminate
+                     when reading STDIN like so:
+                        cat big_input_file | ./mce_script.pl | head -10
+
+                     This works fine without the -setpgrp option:
+                        ./mce_script.pl < big_input_file | head -10
+
+Nothing is exported by default. Exportable are 1 variable and 2 subroutines.
+
+ $tmp_dir          - Path to the temporary directory.
  stop_and_exit     - Described below
  sys_cmd           - Described below
 
@@ -442,9 +540,13 @@ Mario E. Roy, S<E<lt>marioeroy AT gmail DOT comE<gt>>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2012 by Mario E. Roy
+Copyright (C) 2012-2013 by Mario E. Roy
 
-MCE::Signal is free software; you can redistribute it and/or modify it under
-the same terms as Perl itself L<http://dev.perl.org/licenses/>.
+This program is free software; you can redistribute it and/or modify it
+under the terms of either: the GNU General Public License as published
+by the Free Software Foundation; or the Artistic License.
+
+See L<http://dev.perl.org/licenses/> for more information.
 
 =cut
+
